; Author: Peter N Lewis <peter@stairways.com.au>
; Version: 1.0, 7 July 2001
; Copyright 2001 Peter N Lewis
; This code is free for any use.  Please attribute the source appropriately.

.include "1200def.inc"
.cseg
.org 0

.def	rZero	= r0

.def	rLastRowTime = r2
.def	rRow = r3
.def	rBrite = r6
.def	rLastInputBit = r7
.def	rLastChangeTime = r8
.def	rBytes = r9
.def	rBit = r10
.def	rLength = r11
.def	rChecksum = r12
.def	rData = r13
.def	rDataPos = r14

.def	rNewControl = r16
.def	rNewHi = r17
.def	rNewLo = r18

.def	rDigit0 = r20
.def	rDigit1 = r21
.def	rDigit2 = r22
.def	rDigit3 = r23

.def	rTemp1 = r24
.def	rTemp2 = r25
.def	rTemp3 = r26
.def	rDispNegative = r27	; Used only in SetNumber

.def	rParam1 = r28
.def	rParam2 = r29
.def	rParam3 = r30
.def	rParam4 = r31

.equ	kDigitDP = $01
.equ	kDigitA = $20
.equ	kDigitB = $02
.equ	kDigitC = $08
.equ	kDigitD = $40
.equ	kDigitE = $80
.equ	kDigitF = $10
.equ	kDigitG = $04
.equ	kDigitAll = kDigitA + kDigitB + kDigitC + kDigitD + kDigitE + kDigitF + kDigitG

.equ	kDigit0 = kDigitAll - kDigitG
.equ	kDigit1 = kDigitB + kDigitC
.equ	kDigit2 = kDigitAll - kDigitF - kDigitC
.equ	kDigit3 = kDigitAll - kDigitF - kDigitE
.equ	kDigit4 = kDigitAll - kDigitA - kDigitE - kDigitD
.equ	kDigit5 = kDigitAll - kDigitB - kDigitE
.equ	kDigit6 = kDigitAll - kDigitB
.equ	kDigit7 = kDigitA + kDigitB + kDigitC
.equ	kDigit8 = kDigitAll
.equ	kDigit9 = kDigitAll - kDigitE - kDigitD
.equ	kDigitBlank = 0
.equ	kDigitNegative = kDigitG

.equ	kDigitLetterL = kDigitF + kDigitE + kDigitD
.equ	kDigitLetterO = kDigit0
.equ	kDigitLetterS = kDigit5
.equ	kDigitLetterT = kDigit7

.equ	kBitCount = 54	; MHz/8 counts per clock

; Calling conventions:
; parameters (in/out) passed in rParam1, rParam2, ...
; parameters are not preserved
; rTemp1, rTemp2 ... are not preserved
; all other registers are preserved

RESET:
	clr	rZero
	mov	rLastRowTime, rZero
	mov	rRow, rZero
	mov	rBytes, rZero
	dec	rBytes
	
	ldi	rTemp1,$02		; CLK/8
	out	TCCR0,rTemp1
	out	TCNT0, rZero
	
	ldi	rTemp1,$FF
	out	PORTB,rTemp1
	ldi	rTemp1,$FF
	out	DDRB, rTemp1
	
	ldi	rTemp1,$FF
	out	PORTD,rTemp1
	ldi	rTemp1,$65
	out	DDRD, rTemp1

	in	rTemp1, PIND		; D34 = selector.  *3+1 for data pos
	lsr	rTemp1
	lsr	rTemp1
	andi	rTemp1, $06
	mov	rTemp2, rTemp1
	lsr	rTemp2
	add	rTemp1, rTemp2
	inc	rTemp1
;	ldi	rTemp1, 1
;	ldi	rTemp1, 4
;	ldi	rTemp1, 7
;	ldi	rTemp1, 10
	mov	rDataPos, rTemp1
	
	ldi	rDigit3, $FF-kDigitLetterL
	ldi	rDigit2, $FF-kDigitLetterO
	ldi	rDigit1, $FF-kDigitLetterS
	ldi	rDigit0, $FF-kDigitLetterT
	ldi	rTemp1, $7E
	mov	rBrite, rTemp1
	
	wdr
	ldi	rTemp1, $0F
	out	WDTCR, rTemp1
		
LOOP:
	in	rTemp1, PIND
	eor	rTemp1, rLastInputBit
	andi	rTemp1, $02
	breq	FinishInputX
	ldi	rTemp1, $02
	eor	rLastInputBit, rTemp1
	
	mov	rTemp3, rLastChangeTime
	in	rLastChangeTime, TCNT0
	sub	rTemp3, rLastChangeTime
	neg	rTemp3
	
	tst	rBytes
	brpl	Receiving
	
	mov	rTemp1, rLastInputBit
	andi	rTemp1, $02
	brne	FinishInputX
	
	cpi	rTemp3, kBitCount*3
	brlo	FinishInputX
	
	mov	rBytes, rZero
	ldi	rTemp1, 8
	mov	rBit, rTemp1
	ldi	rTemp1, $55
	mov	rChecksum, rTemp1
	
FinishInputX:
	rjmp	FinishInput
	
Receiving:
	cpi	rTemp3, kBitCount/2
	brlo	FailPacket
	cpi	rTemp3, kBitCount*3/2
	brlo	Got0
	cpi	rTemp3, kBitCount*3
	brsh	FailPacket
Got1:	sec
	rjmp	GotBit
Got0:	clc
GotBit: rol	rData
	dec	rBit
	brne	FinishInput
	
	ldi	rTemp1, 8
	mov	rBit, rTemp1		
	inc	rBytes
	add	rChecksum, rData
	
	mov	rTemp1, rBytes
	cpi	rTemp1, 1
	brne	NotLenByte
	
	mov	rLength, rData
	ldi	rTemp1, 1
	add	rTemp1, rDataPos
	cp	rTemp1, rLength
	brsh	FailPacket

	rjmp	FinishInput
	
NotLenByte:
	mov	rTemp1, rDataPos
	sub	rTemp1, rBytes
	inc	rTemp1
	breq	Byte1
	inc	rTemp1
	breq	Byte2
	inc	rTemp1
	breq	Byte3
	mov	rTemp1, rBytes
	sub	rTemp1, rLength
	cpi	rTemp1, 2
	brne	FinishInput
; Finished

	tst	rChecksum
	brne	FailPacket

	; Turn off leds to avoid bright flicker at low brightness	
	ldi	rTemp1,$65
	out	PORTD, rTemp1

	wdr
	rcall	SetNumber
	rjmp	FailPacket

Byte1:	mov	rNewControl, rData
	rjmp	FinishInput
	
Byte2:	mov	rNewLo, rData
	rjmp	FinishInput
	
Byte3:	mov	rNewHi, rData
	rjmp	FinishInput
		 
FailPacket:
	mov	rBytes, rZero
	dec	rBytes

FinishInput:
	in	rTemp1, TCNT0
	sub	rTemp1, rLastRowTime

	mov	rTemp2, rBrite
	tst	rTemp2
	brpl	NotOff
	neg	rTemp2
NotOff:
	cpi	rTemp2, 6
	brsh	Dont6
	tst	rTemp2
	breq	Dont6
	ldi	rTemp2, 6
Dont6:	cp	rTemp1, rTemp2
	brsh	Continue1
	rjmp	NotRowTime
Continue1:
	add	rLastRowTime, rTemp1
	
	ldi	rTemp1,$65
	out	PORTD, rTemp1
	
; if rRow = 0 then begin
	cp	rRow, rZero
	brne	NotRowZero

; Brightness Sequence:
; 126:  126   126   126   126
; 124:  124    -4   124    -4
; 122:  122    -6   122    -6
;   6:    6  -122     6  -122
;   4:    4  -124   -65     4  -124   -65
;   2:    2  -126  -125  -127     2  ...
;   0: -128  -128  -128  -128

; 1 to 127 on time
; -128 to -1 off time (negative)	
	mov	rTemp1, rBrite
	mov	rTemp2, rTemp1
	cpi	rTemp1, 126
	breq	NoBrightChange
	cpi	rTemp1, -128
	breq	NoBrightChange
	ldi	rTemp1, -65
	cpi	rTemp2, -124
	breq	FinishBrightChange
	ldi	rTemp1, 4
	cpi	rTemp2, -65
	breq	FinishBrightChange
	ldi	rTemp1, -125
	cpi	rTemp2, -126
	breq	FinishBrightChange
	ldi	rTemp1, -127
	cpi	rTemp2, -125
	breq	FinishBrightChange
	ldi	rTemp1, 2
	cpi	rTemp2, -127
	breq	FinishBrightChange
	mov	rTemp1, rTemp2
	subi	rTemp1, 128
FinishBrightChange:
	mov	rBrite, rTemp1
NoBrightChange:

NotRowZero:
	mov	rTemp3, rRow
	cpi	rTemp3, 1
	brlo	Row0
	breq	Row1
	cpi	rTemp3, 2
	breq	Row2
Row3:	out	PORTB, rDigit3
	ldi	rTemp1, $64
	rjmp	SetRow
Row2:	out	PORTB, rDigit2
	ldi	rTemp1, $61
	rjmp	SetRow
Row1:	out	PORTB, rDigit1
	ldi	rTemp1, $25
	rjmp	SetRow
Row0:	out	PORTB, rDigit0
	ldi	rTemp1, $45
SetRow:	tst	rBrite
	brmi	NoDisplay
	breq	NoDisplay
	out	PORTD, rTemp1
NoDisplay:

	inc	rTemp3
	andi	rTemp3, $03
	mov	rRow, rTemp3
	
NotRowTime:
	rjmp	LOOP

; rNewLo, rNewHi = number, rNewControl = control
SetNumber:
	mov	rTemp1, rNewControl
	andi	rTemp1, $FC
	eor	rNewControl, rTemp1
	lsr	rTemp1
	mov	rBrite, rTemp1
	
	ldi	rTemp1, $FF-kDigitBlank
	mov	rDispNegative, rTemp1
	cp	rNewHi, rZero
	brge	SetNumberPositive
	ldi	rTemp1, $FF-kDigitNegative
	mov	rDispNegative, rTemp1
	com	rNewLo
	com	rNewHi
	sec
	adc	rNewLo, rZero
	adc	rNewHi, rZero
SetNumberPositive:
	ldi	rDigit3, $FF-kDigitBlank
	ldi	rDigit2, $FF-kDigitBlank
	ldi	rDigit1, $FF-kDigitBlank
	mov	rParam1, rNewLo
	mov	rParam2, rNewHi
	rcall	DivMod10
	rcall	ConvertToDigitP3
	mov	rDigit0, rParam3
	cp	rParam1, rZero
	brne	SNNotZero1
	cp	rParam2, rZero
	brne	SNNotZero1
	cp	rNewControl, rZero
	brne	SNNotZero1
	mov	rDigit1, rDispNegative
	ret
SNNotZero1:
	rcall	DivMod10
	rcall	ConvertToDigitP3
	cpi	rNewControl, 1
	brne	NotDecimal1
	andi	rParam3, $FF-kDigitDP
NotDecimal1:
	mov	rDigit1, rParam3
	cp	rParam1, rZero
	brne	SNNotZero2
	cpi	rNewControl, 2
	brsh	SNNotZero2
	mov	rDigit2, rDispNegative
	ret
SNNotZero2:
	rcall	DivMod10
	rcall	ConvertToDigitP3
	cpi	rNewControl, 2
	brne	NotDecimal2
	andi	rParam3, $FF-kDigitDP
NotDecimal2:
	mov	rDigit2, rParam3
	cp	rParam1, rZero
	brne	SNNotZero3
	cpi	rNewControl, 3
	brsh	SNNotZero3
	mov	rDigit3, rDispNegative
	ret
SNNotZero3:
	rcall	DivMod10
	rcall	ConvertToDigitP3
	cpi	rNewControl, 3
	brne	NotDecimal3
	andi	rParam3, $FF-kDigitDP
NotDecimal3:
	mov	rDigit3, rParam3
	ret

; input rParam3 = 0-9, output, rParam3 = digit
ConvertToDigitP3:
	cpi	rParam3, 3
	breq	Digit3
	brlo	DigitSmall
	cpi	rParam3, 7
	breq	Digit7
	brsh	DigitBig
	cpi	rParam3, 5
	brlo	Digit4
	breq	Digit5
	rjmp	Digit6

DigitBig:
	cpi	rParam3, 8
	breq	Digit8
	rjmp	Digit9
	
DigitSmall:
	cpi	rParam3, 1
	breq	Digit1
	brsh	Digit2
Digit0:	ldi	rParam3, $FF - kDigit0
	ret
Digit1:	ldi	rParam3, $FF - kDigit1
	ret
Digit2:	ldi	rParam3, $FF - kDigit2
	ret
Digit3:	ldi	rParam3, $FF - kDigit3
	ret
Digit4:	ldi	rParam3, $FF - kDigit4
	ret
Digit5:	ldi	rParam3, $FF - kDigit5
	ret
Digit6:	ldi	rParam3, $FF - kDigit6
	ret
Digit7:	ldi	rParam3, $FF - kDigit7
	ret
Digit8:	ldi	rParam3, $FF - kDigit8
	ret
Digit9:	ldi	rParam3, $FF - kDigit9
	ret

; Wonderous code from John West - THANKS!
; rParam1 = Lo, rParam2 = Hi Assert(HiLo >= 0).  Return also rParam3 = remainder
DivMod10:
	mov	rTemp1, rParam1
	mov	rTemp2, rParam2
	lsr	rTemp2
	ror	rTemp1
	lsr	rTemp2
	ror	rTemp1
	add	rTemp1, rParam1
	adc	rTemp2, rParam2
	lsr	rTemp2
	ror	rTemp1
	add	rTemp1, rParam1
	adc	rTemp2, rParam2
	lsr	rTemp2
	ror	rTemp1
	lsr	rTemp2
	ror	rTemp1
	lsr	rTemp2
	ror	rTemp1
	add	rTemp1, rParam1
	adc	rTemp2, rParam2
	lsr	rTemp2
	ror	rTemp1
	add	rTemp1, rParam1
	adc	rTemp2, rParam2
	lsr	rTemp2
	ror	rTemp1
	lsr	rTemp2
	ror	rTemp1
	lsr	rTemp2
	ror	rTemp1
	add	rTemp1, rParam1
	adc	rTemp2, rParam2
	lsr	rTemp2
	ror	rTemp1
	add	rTemp1, rParam1
	adc	rTemp2, rParam2
	lsr	rTemp2
	ror	rTemp1
	mov	rParam3, rTemp1
	andi	rParam3, 248
	lsr	rTemp2
	ror	rTemp1
	lsr	rTemp2
	ror	rTemp1
	add	rParam3, rTemp1
	andi	rParam3, 254
	lsr	rTemp2
	ror	rTemp1
	neg	rParam3
	add	rParam3, rParam1
	mov	rParam1, rTemp1
	mov	rParam2, rTemp2
	ret
	
